
#define NODE0 (t_my_ref)0
#define NODE1 (t_my_ref)1

#include "dynare2.h"
#include "d_tab.hh"
#include "moments.h"
#include "dynare_params.h"
#include "../sylv/cc/SylvException.h"
#include "../tl/cc/tl_exception.h"
#include "../tl/cc/t_container.h"
#include "../tl/cc/sparse_tensor.h"
#include "../tl/cc/sthread.h"
#include "../kord/decision_rule.h"
#include "../kord/korder.h"
#include "../kord/first_order.h"

//void first_order(void);
void constant(vec_double &);
void k_order(int, int);
void disp_dr(int);
void init_matrices(int);
void correct(int);
void decision_rules(const int,const char *,const char*);

// global variables

type_model model;
type_symbol_table symbols;
type_variable_table variables;
type_constants constants;
t_var_deriv_index var_deriv_index;
t_h_deriv deriv_col;
mtl::dense1D<s_mixed> mixed;
struct model_struct sm;
vec_double ys, yss, exe, param;
MatA y, ex;
dmatx f;
MatA zx;
MatA zxu;
tm_it end_resid, end_endo_deriv, end_first_deriv;
int deriv_order;
TensorContainer<SparseTensor> md(1);

extern Moments moments;

bool operator == (const struct s_node & s1, const struct s_node &s2){
  //  std::cout << s1.arg1 << " " << s1.arg2 << " " << s1.arg2 << " " << s2.arg2 << " " << s1.code << " " << s2.code << "\n";
  return (s1.arg1 == s2.arg1 && s1.arg2==s2.arg2 && s1.code == s2.code);
}

bool operator !=(const struct s_node & s1, const struct s_node &s2){
  return (s1.arg1 != s2.arg1 || s1.arg2!=s2.arg2 || s1.code != s2.code);
}

bool variable_comp::operator()(const struct s_variable_key & s1, const struct s_variable_key & s2)
{
  if (s1.lag < s2.lag)
    {
      return 1;
    }
  else if (s1.lag > s2.lag)
    {
      return 0;
    }
  else
    {
      return strcmp(s1.name,s2.name) < 0; 
    }
}

t_my_ref add_node_check(int code, t_my_ref arg1, t_my_ref arg2)
  /* Code contains the code of the operator
     arg1 contains the address of the first passed argument
     arg2 contains the address of the second passed argument
     if argi=0 then result=NODE0
     if argi=1 then result=NODE1
  */
{
#ifdef __MINGW32__
	if (code == ACOSH || code == ASINH || code == ATANH) {
		fprintf(stderr, "cosh, sinh, and tanh not implemented for MinGW\n");
		exit(1);
	}
#endif

  switch(code)
    {
    case PLUS:
      if (arg1 != NODE0 && arg2 != NODE0)
	{
	  return add_node(PLUS,arg1,arg2);
	}
      else if (arg1 != NODE0)
	{
	  return arg1;
	}
      else if (arg2 != NODE0)
	{
	  return arg2;
	}
      else
	{
	  return NODE0;
	}
      break;
    case MINUS:
      if (arg1 != NODE0 && arg2 != NODE0)
	{
	  return add_node(MINUS,arg1,arg2);
	}
      else if (arg1 != NODE0)
	{
	  return arg1;
	}
      else if (arg2 != NODE0)
	{
	  return add_node(MINUS,0,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case TIMES:
      if (arg1 != NODE0 && arg1 != NODE1 && arg2 != NODE0 && arg2 != NODE1)
	{
	  return add_node(TIMES,arg1,arg2);
	}
      else if (arg1 != NODE0 && arg1 != NODE1 && arg2 == NODE1)
	{
	  return arg1;
	}
      else if (arg2 != NODE0 && arg2 != NODE1 && arg1 == NODE1)
	{
	  return arg2;
	}
      else if (arg2 == NODE1 && arg1 == NODE1)
	{
	  return NODE1;
	}
      else
	{
	  return NODE0;
	}
      break;
    case DIVIDE:
      if (arg1 != NODE0 && arg2 != NODE0 && arg2 != NODE1)
	{
	  return add_node(DIVIDE,arg1,arg2);
	}
      else if (arg2 == NODE1)
	{
	  return arg1;
	}
      else if (arg1 == NODE0 && arg2 != NODE0)
	{
	  return NODE0;
	}
      else
	{
	  std::cerr << "DIVIDE 0/0 non available\n";
	  exit(-1);
	}
      break;
    case EXP:
      if (arg2 != NODE0)
	{
	  return add_node(EXP,arg1,arg2);
	}
      else
	{
	  return NODE1;
	}
      break;
    case LOG:
      if (arg2 != NODE0)
	{
	  return add_node(LOG,arg1,arg2);
	}
      else
	{
	  std::cerr << "log(0) isn't available\n";
	  exit(-1);
	}
      break;
    case LOG10:
      if (arg2 != NODE0)
	{
	  return add_node(LOG10,arg1,arg2);
	}
      else
	{
	  std::cerr << "log10(0) isn't available\n";
	  exit(-1);
	}
      break;
    case COS:
      if (arg2 != NODE0)
	{
	  return add_node(COS,arg1,arg2);
	}
      else
	{
	  return NODE1;
	}
      break;
    case SIN:
      if (arg2 != NODE0)
	{
	  return add_node(SIN,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case TAN:
      if (arg2 != NODE0)
	{
	  return add_node(TAN,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case ACOS:
	  return add_node(ACOS,arg1,arg2);
      break;
    case ASIN:
      if (arg2 != NODE0)
	{
	  return add_node(ASIN,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case ATAN:
      if (arg2 != NODE0)
	{
	  return add_node(ATAN,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case COSH:
      if (arg2 != NODE0)
	{
	  return add_node(COSH,arg1,arg2);
	}
      else
	{
	  return NODE1;
	}
      break;
    case SINH:
      if (arg2 != NODE0)
	{
	  return add_node(SINH,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case TANH:
      if (arg2 != NODE0)
	{
	  return add_node(TANH,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case ACOSH:
	  return add_node(ACOSH,arg1,arg2);
      break;
    case ASINH:
      if (arg2 != NODE0)
	{
	  return add_node(ASINH,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case ATANH:
      if (arg2 != NODE0)
	{
	  return add_node(ATANH,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case SQRT:
      if (arg2 != NODE0)
	{
	  return add_node(SQRT,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case POWER:
      if (arg1 != NODE0 && arg2 != NODE0 && arg2 != NODE1)
	{
	  return add_node(POWER,arg1,arg2);
	}
      else if (arg2 == NODE1)
	{
	  return arg1;
	}
      else if (arg2 == NODE0)
	{
	  return NODE1;
	}
      else
	{
	  return NODE0;
	}
      break;
    default:
      std::cerr << "missing code in add_node_check " << code << "\n";
      exit(-1);
    }
}

t_my_ref add_node(int code, t_my_ref arg1, t_my_ref arg2)
{
  struct s_node node;

  /* for commutative operators, argument addresses are ordered */
    if ((code == PLUS || code == TIMES) && arg1 > arg2)
    {
      t_my_ref temp=arg2;
      arg2=arg1;
      arg1=temp;
    }

  node.code = code;
  node.arg1 = arg1;
  node.arg2 = arg2;
  if (arg1 != NODE0 && arg1 != NODE1) ++(arg1->ref);
  if (arg2 != NODE0 && arg2 != NODE1) ++(arg2->ref);

  tm_it address = find(model.begin(),model.end(),node);

  if (address == model.end() )
    {
      node.ref = 0;

      model.push_back(node);
      //      return (--model.end())._M_node;
      return &model.back();
    }
  else
    {
      return &(*address);
    }
}

void add_result(int arg1, t_my_ref arg2)
{
  if (arg2 != NODE0 && arg2 != NODE1) ++(arg2->ref);

  struct s_node node;

  node.code = EQUAL_SIGN;
  node.arg1 = NODE0;
  node.arg2 = arg2;
  node.ref = arg1;
  node.no = 0;
  model.push_back(node);
  if (arg1 == sm.endo_nbr-1)
    {
      end_resid = model.end();
      end_resid--;
    }
}

void add_deriv(int arg1, t_my_ref arg2, int arg3, int arg4, int arg5, const IntSequence & intseq)
{
  if (arg2 != NODE0 && arg2 != NODE1) ++(arg2->ref);

  struct s_node node;

  node.code = EQUAL_SIGN;
  node.arg1 = NODE0;
  node.arg2 = arg2;
  node.ref = arg1;

  static int index;
  node.no = ++index; 
  s_deriv deriv(arg4, arg3, arg5, intseq);
  deriv_col.insert(t_h_deriv::value_type(index, deriv));

  model.push_back(node);
}

void add_symbol(const char *name, int type)
{
  symbols[name].type=type;
  if (type == 0)
    {
      symbols[name].nbr=sm.exo_nbr;
      sm.exo_nbr++;
    }
  else if (type == 1)
    {
      symbols[name].nbr=sm.endo_nbr;
      sm.endo_nbr++;
    }
  else if (type == 3)
    {
      symbols[name].nbr=sm.param_nbr;
      sm.param_nbr++;
    }
}

void add_symbol_range(char *name1, char *name2, int type)
{
  int i1, i2, i3;
  char name[200],*p1,*p2;

  p1=name1;
  p2=name2;
  i1=0;
  while (*p1 == *p2)
    {
      p1++;
      p2++;
      i1++;
    }

  if (*p1 >= '0' & *p1 <= '9')
    {
      if (sscanf(p1,"%d",&i2) != 1)
	{
	  fprintf(stderr,"VAR, VAREXO or VARDEF: error in name range expression\n");
	  return;
	}    
      if (sscanf(p2,"%d",&i3) != 1)
	{
	  fprintf(stderr,"VAR, VAREXO or VARDEF: error in name range expression\n");
	  return;
	}    
      if (i3 < i2)
	{
	  std::cerr << "VAR, VAREXO or VARDEF: error in name range expression\n";
	  return;
	}
    
      strncpy(name,name1,i1);
      while (i2 <= i3)
	{
	  sprintf(name+i1,"%d\0",i2);
	  add_symbol(name, type);
	  i2++;
	}
    }
  else
    {
      std::cerr << "VAR, VAREXO or VARDEF: error in name range expression\n";
      return;
    }
  return;
}

t_my_ref add_constant(int code, const char *s)
{
  if (strcmp(s,"0")==0) return NODE0;
  if (strcmp(s,"1")==0) return NODE1;

  s_variable_key v;
  v.name = s;
  v.lag = 0;
  type_variable_table::iterator address = variables.find(v);

  if (address == variables.end() )
    {
      add_symbol(s,3);
      set_constant(s);
      return add_variable(s,0,0);
    }
  return variables[v].address;
}


t_my_ref add_variable(const char *s,int lag, int dummy)
{
  type_symbol_table::iterator i = symbols.find(s);
  if (i == symbols.end())
    {
      std::cerr << "Unknown symbol in model: " << s;
      exit(1);
    }

  if (i->second.type == 0 && lag != 0)
    {
      std::cerr << "Exogenous shocks must be in current period. Use another endogenous variable.\n";
	}
  s_variable_key v;
  v.name = i->first;
  v.lag = lag;
  if (lag < sm.maxlag)
    {
      sm.maxlag = lag;
    }
  else if (lag > sm.maxlead)
    {
      sm.maxlead = lag;
    }

  type_variable_table::iterator address = variables.find(v);

  if (address == variables.end() )
    {
      variables[v].type = i->second.type;
      if (i->second.type == 1)
	{
	  sm.dyn_endo_nbr++;
	}
      variables[v].nbr = i->second.nbr;
      if (dummy == 0)
	{
	  s_node node;
	  node.code = 0;
	  node.arg1 = NODE0;
	  node.arg2 = NODE0;
	  model.push_back(node);
	  //	  variables[v].address = (--model.end())._M_node;
	  variables[v].address = &(model.back());
	}
      else
	{
	  variables[v].address = (t_my_ref)0;
	}
    }
  return variables[v].address;
}

void derivator( int deriv_order )
{
  std::cout << "derivatives\n";

  // initialize derivatives of variables 
  int nbr_deriv_var = 0;
  for (t_var_deriv_index::iterator i=var_deriv_index.begin(); i != var_deriv_index.end(); i++)
    {
      if ( ((*i)->second).address != (t_my_ref)0 )
	{
	  //	  ((tm_it)(((*i)->second).address))->derivatives[nbr_deriv_var] = NODE1;
	  ((((*i)->second).address))->derivatives[nbr_deriv_var] = NODE1;
	}
      ++nbr_deriv_var;
    }

  tm_it k2 = model.begin();
  --k2;
  vec_int deriv_block(nbr_deriv_var);
  deriv_block[0] = 0;
  for (int i=1; i < nbr_deriv_var; i++)
    {
      deriv_block[i] = 1;
    }
    
  // loop on order of derivation
  for(int i=0; i < deriv_order; i++)
    {
      std::cout << "  order " << i+1 << "\n";
      // updates nbr of derivatives to skip
      tm_it k1 = k2;
      ++k1;
      k2=model.end();
      --k2;
      // loop on variables of derivation
      int offset = 0;
      for (int j=0; j < nbr_deriv_var; j++)
	{
	  std::cout << "    var " << j << "\n";
	  if (i > 0)
	    {
	      deriv_block[j]=deriv_block[j]*(nbr_deriv_var+i-j-1)/i;
	      offset += deriv_block[j];
	    }
	  // loop on nodes
	  for (tm_it k=k1;; k++)
	    {
	      if (k->code == 0) continue; // this is a leaf

	      t_my_ref a, d1, d2, t1, t2, t3, t4, t5;
	      a = k->arg1;
	      if (a == NODE0 || a == NODE1) d1 = NODE0;
	      else
		{
		  //		  t_derivatives dd = ((tm_it)a)->derivatives;
		  t_derivatives dd = a->derivatives;
		  t_derivatives::iterator id = dd.find(j);
		  if (id == dd.end()) d1 = NODE0;
		  else d1 = dd[j];
		}
	      a = k->arg2;
	      if (a == NODE0 || a == NODE1) d2 = NODE0;
	      else
		{
		  //		  t_derivatives dd = ((tm_it)(k->arg2))->derivatives;
		  t_derivatives dd = (k->arg2)->derivatives;
		  t_derivatives::iterator id = dd.find(j);
		  if (id == dd.end()) d2 = NODE0;
		  else d2 = dd[j];
		} 
	      switch (k->code)
		{
		case PLUS:
		  if (d1 != NODE0 || d2 != NODE0)
		    {
		      k->derivatives[j]=add_node_check(PLUS,d1,d2);
		    }
		  break;
		case MINUS:
		  if (d1 != NODE0 || d2 != NODE0)
		    {
		      k->derivatives[j]=add_node_check(MINUS,d1,d2);
		    }
		  break;
		case TIMES:
		  if (d1 != NODE0 || d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,d1,k->arg2);
		      t2 = add_node_check(TIMES,d2,k->arg1);
		      k->derivatives[j]=add_node_check(PLUS,t1,t2);
		    }
		  break;
		case DIVIDE:
		  if (d1 != NODE0 || d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,d1,k->arg2);
		      t2 = add_node_check(TIMES,d2,k->arg1);
		      t3 = add_node_check(MINUS,t1,t2);
		      t4 = add_node_check(TIMES,k->arg2,k->arg2);
		      k->derivatives[j]=add_node_check(DIVIDE,t3,t4);
		    }
		  break;
		case EXP:
		  if (d2 != NODE0)
		    {
		      k->derivatives[j]=add_node_check(TIMES,d2,&(*k));
		    }
		  break;

		case LOG:
		  if (d2 != NODE0)
		    {
		      k->derivatives[j]=add_node_check(DIVIDE,d2,k->arg2);
		    }
		  break;
		case LOG10:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(EXP,NODE1,NODE1);
		      t2 = add_node_check(LOG10,NODE1,t1);
		      t3 = add_node_check(DIVIDE,d2,k->arg2);
		      k->derivatives[j]=add_node_check(TIMES,t2,t3);
		    }
		  break;
		case COS:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(SIN,NODE0,k->arg2);
		      t2 = add_node_check(MINUS,NODE0,t1);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case SIN:
		  if (d2 != NODE0)
		    {
		      t2 = add_node_check(COS,k->arg1,k->arg2);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case TAN:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,&(*k),&(*k));
		      t2 = add_node_check(PLUS,NODE1,t1);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case ACOS:
		  if (d2 != NODE0)
		    {
		  t1 = add_node_check(SIN,NODE0,&(*k));
		  t2 = add_node_check(DIVIDE,d2,t1);
		  k->derivatives[j]=add_node_check(MINUS,NODE0,t2);
		    }
		  break;
		case ASIN:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(COS,NODE0,&(*k));
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t1);
		    }
		  break;
		case ATAN:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,k->arg2,k->arg2);
		      t2 = add_node_check(PLUS,NODE1,t1);
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t2);
		    }
		  break;
		case COSH:
		  if (d2 != NODE0)
		    {
		      t2 = add_node_check(SINH,k->arg1,k->arg2);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case SINH:
		  if (d2 != NODE0)
		    {
		      t2 = add_node_check(COSH,k->arg1,k->arg2);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case TANH:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,&(*k),&(*k));
		      t2 = add_node_check(MINUS,NODE1,t1);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case ACOSH:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(SINH,NODE0,&(*k));
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t1);
		    }
		  break;
		case ASINH:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(COSH,NODE0,&(*k));
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t1);
		    }
		  break;
		case ATANH:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,k->arg2,k->arg2);
		      t2 = add_node_check(MINUS,NODE1,t1);
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t2);
		    }
		  break;
		case SQRT:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(PLUS,&(*k),&(*k));
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t1);
		    }
		  break;
		case POWER:
		  if (d1 != NODE0)
		    {
		      if (d2==0)
			{
			  t1 = add_node_check(MINUS,k->arg2,NODE1);
			  t2 = add_node_check(POWER,k->arg1,t1);
			  t3 = add_node_check(TIMES,k->arg2,t2);
			  k->derivatives[j]=add_node_check(TIMES,d1,t3);
			}
		      else
			{
			  t1 = add_node_check(LOG,NODE0,k->arg1);
			  t2 = add_node_check(TIMES,d2,t1);
			  t3 = add_node_check(TIMES,d1,k->arg2);
			  t4 = add_node_check(DIVIDE,t3,k->arg1);
			  t5 = add_node_check(PLUS,t2,t4);
			  k->derivatives[j]=add_node_check(TIMES,t5,&(*k));
			}
		    }
		  break;
		case EQUAL_SIGN:
		  // skip symmetric derivatives
		  if (i == 0 && d2 != NODE0)
		    {
		      IntSequence intseq(1);
		      intseq[0] = j;
		      add_deriv(k->ref,d2,j+1,j,i,intseq);
		    }
		  else {
			  t_h_deriv::const_iterator it = deriv_col.find(k->no);
			  if (it != deriv_col.end() && (*it).second.var >= j && d2 != NODE0)
		    {
		      IntSequence intseq(i+1);
		      for(int is=0;is<i;++is) intseq[is] = (*it).second.intseq[is];
		      intseq[i] = j;
		      add_deriv(k->ref,d2,offset+(*it).second.col,j,i,intseq);
		    }
		  }
		  break;
		default:
		  std::cout << "Missing code in DERIV:" << k->code;
		}                    /* end switch */
	      if (k == k2) break;
	    }                        /* end for k=k1..k2 */
	  if (i == 0 && j == sm.dyn_endo_nbr - 1)
	    {
	      end_endo_deriv = model.end();
	      --end_endo_deriv;
	    }
	}                            /* end for j=0...nbr_deriv_var */

      for(tm_rit k = model.rbegin(); k != (tm_rit)k2; k++)
	{
	  if (k->code != 0 && k->code != EQUAL_SIGN && k->ref == 0){
	    //	    if (k->arg1 != NODE0 && k->arg1 != NODE1) --(((tm_it)k->arg1)->ref);
	    //	    if (k->arg2 != NODE0 && k->arg2 != NODE1) --(((tm_it)k->arg2)->ref);
	    if (k->arg1 != NODE0 && k->arg1 != NODE1) --((k->arg1)->ref);
	    if (k->arg2 != NODE0 && k->arg2 != NODE1) --((k->arg2)->ref);
	  }
	}
      tm_it k3 = model.end();
      for(tm_it k = k2; k != k3;)
	{
	  tm_it k1 = k;
	  ++k1;
	  if (k->code != 0 && k->code != EQUAL_SIGN && k->ref == 0){
       	    model.erase(k);
	  }
	  k=k1;
	}
      if ( i == 0)
	{
	  end_first_deriv = model.end();
	  --end_first_deriv;
	}
    }                                /* end for i=0...deriv_order */
}

void init_var()
{
  ys.resize(sm.endo_nbr);
}
  
void init_varexo()
{
  exe.resize(sm.exo_nbr);
}
  
void init_param()
{
  param.resize(sm.param_nbr);
}

class DynareNameList : public NameList {
	vector<const char*> names;
public:
	DynareNameList();
	int getNum() const
		{return (int)names.size();}
	const char* getName(int i) const
		{return names[i];}
};

DynareNameList::DynareNameList()
{
	for (t_var_deriv_index::iterator i = var_deriv_index.begin();
		 i != var_deriv_index.end(); i++) {
		if (((*i)->first).lag == 0 && ((*i)->second).type == 1)
			names.push_back(((*i)->first).name);
    }
}

class DynareStateNameList : public NameList {
	vector<const char*> names;
public:
	DynareStateNameList(const DynareNameList& dnl, int nstat, int npred, int nboth, int nexog);
	int getNum() const
		{return (int)names.size();}
	const char* getName(int i) const
		{return names[i];}
};

DynareStateNameList::DynareStateNameList(const DynareNameList& dnl, int nstat, int npred,
										 int nboth, int nexog)
{
	// add predetermined variables
	for (int i = nstat; i < nstat+npred+nboth; i++)
		names.push_back(dnl.getName(i));

	// add exogenous variables
	for (t_var_deriv_index::iterator i = var_deriv_index.begin();
		 i != var_deriv_index.end(); i++) {
		if (((*i)->first).lag == 0 && ((*i)->second).type == 0)
			names.push_back(((*i)->first).name);
	}
}

class SimulationWorker : public THREAD {
	SimResults& res;
	const DecisionRule<KOrder::fold>& dr;
	DecisionRule<KOrder::fold>::emethod em;
	int np;
	const Vector& st;
	const ShockRealization& sr;
public:
	SimulationWorker(SimResults& sim_res,
					 const DecisionRule<KOrder::fold>& dec_rule,
					 DecisionRule<KOrder::fold>::emethod emet, int num_per,
					 const Vector& start, const ShockRealization& shock_r)
		: res(sim_res), dr(dec_rule), em(emet), np(num_per), st(start), sr(shock_r) {}
	void operator()();
};

void SimulationWorker::operator()()
{
	const TwoDMatrix* m = dr.simulate(em, np, st, sr);
	{
		SYNCHRO syn(&res, "simulation");
		res.addDataSet(m);
	}
}

int main(int argc, char** argv)
{
	DynareParams params(argc, argv);
	if (params.help) {
		params.printHelp();
		return 0;
	}
	THREAD_GROUP::max_parallel_threads = params.num_threads;

	extern FILE *yyin;

	extern int yydebug;
	//  yydebug = 1;
	  

	add_constant(INUMBER,"0");
	add_constant(INUMBER,"1");

	yyin=fopen(params.modname, "r");
	yyparse();

	sort_deriv_variables();

	derivator(deriv_order);

	// set y partitioning 
	int nstat = sm.static_nbr;
	int nboth = sm.c_both_nbr;
	int npred = sm.s_lagged_nbr - sm.c_both_nbr;
	int nforw = sm.fwrd_nbr - sm.c_both_nbr;
	int nexog = sm.exo_nbr; 

	DynareNameList dnl;
	DynareStateNameList dsnl(dnl, nstat, nboth, npred, nexog);

	try {
		tls.init(deriv_order, sm.dyn_endo_nbr+sm.exo_nbr);

		// Michel's stuff
		y = MatA(-sm.maxlag+sm.maxlead+1,sm.endo_nbr);
		ex = MatA(-sm.maxlag+sm.maxlead+1,sm.exo_nbr);

		copy(exe,rows(ex)[1]);

		std::cout << "nlsolve\n";
		nlsolve(ys);
		std::cout << "Steady state\n";
		print_vector(ys);

		for(int i = 0; i < -sm.maxlag+sm.maxlead+1; i++)
		{
			copy(ys,rows(y)[i]);
		}
		int it = 1-sm.maxlag;
		vec_double zz(sm.endo_nbr);
		vec_double g_v;
		vec_int g_c, g_r;
		f.resize(deriv_order+1);
		execute_tree(md, zz, g_v, g_c, g_r, model.end(), it, 1);
		std::cout << "OK execute\n";
		// end of Michel's stuff

		// make journal name
		char jname[500];
		sprintf(jname, "%s.jnl", params.basename);
		Journal journal(jname);

		// write information on variables to journal
		{
			JournalRecord jr(journal);
			jr << "The model has " << nstat << " static, " << nboth+npred
			   << " predetermined, " << nboth+nforw << " frwd-looking, including "
			   << nboth << " at both times" << endrec;
		}

		// do the first order
		FirstOrder fo(nstat, npred, nboth, nforw, nexog,
					  *(md.get(Symmetry(1))), journal);

		// obtain the covariance matrix of shocks
		TwoDMatrix vcov(nexog, nexog, moments.getVcov().data());

		// do all other orders
		KOrder kord(nstat, npred, nboth, nforw, md,
					fo.getGy(), fo.getGu(), vcov, journal);
		kord.switchToFolded();
		for (int k = 2; k <= md.getMaxDim(); k++) {
			kord.performStep<KOrder::fold>(k);
			if (params.check)
				kord.check<KOrder::fold>(k);
		}

		// set steady state
		Vector steady(kord.getPartY().ny());
		for (int i = 0; i < steady.length(); i++)
			steady[i] = ys[i];

		// open mat file
		char matfile[100];
		sprintf(matfile, "%s.mat", params.basename);
		FILE* matfd;
		if (NULL == (matfd=fopen(matfile, "wb"))) {
			fprintf(stderr, "Couldn't open %s for writing.\n", matfile);
			exit(1);
		}

		// construct the decision rule and simulate
		DecisionRule<KOrder::fold> dr(kord.getFoldDers(), kord.getPartY(), nexog, steady, 1.0);
		dr.writeMat4(matfd, params.prefix);
		RandomGenerator::initSeed(params.seed);
		SimResults res(dnl, dsnl, params.num_per);
		RandomShockRealization sr(vcov);
		{
			JournalRecordPair paa(journal);
			paa << "Performing " << params.num_sim << " stochastic simulations for "
				<< params.num_per << " periods" << endrec;
			THREAD_GROUP gr;
			for (int i = 0; i < params.num_sim; i++) {
				THREAD* worker = new
					SimulationWorker(res, dr, DecisionRule<KOrder::fold>::horner,
									 params.num_per, steady, sr);
				gr.insert(worker);
			}
			gr.run();
			res.writeMat4(matfd, params.prefix);
		}

		fclose(matfd);

	} catch (const TLException& e) {
		printf("Caugth TL exception: ");
		e.print();
	} catch (SylvException& e) {
		printf("Caught Sylv exception: ");
		e.printMessage();
	}

  return 0;
}
