#include "dynare2.h"
#include "d_tab.hh"
#include "../tl/cc/fs_tensor.h"

#define NODE0 (t_my_ref)0
#define NODE1 (t_my_ref)1

extern model_struct sm;
extern dmatx f;
extern MatA ex, y;
extern vec_double param;
extern type_variable_table variables;
extern type_model model;
extern t_h_deriv deriv_col;
#include "../tl/cc/t_container.h"
#include "../tl/cc/sparse_tensor.h"
  

void execute_tree(TensorContainer<SparseTensor> &md, vec_double & zz,vec_double & g_v, vec_int & g_i, vec_int &g_r, tm_it end_node, int it, int flag)
{
	SparseTensor *md_st = NULL;
	if (flag) {
  md_st = new SparseTensor(1,sm.dyn_endo_nbr+sm.exo_nbr,sm.endo_nbr);
  md.insert(md_st);
	}
  // initializes variables for C++ function
  for( type_variable_table::iterator j=variables.begin(); j != variables.end();j++)
    {
      struct s_variable_data s;
      s = j->second;
      switch ( s.type)
	{
	case 0:
	  //	  ((tm_it)s.address)->result = ex(it+(j->first).lag-1,s.nbr);
	  (s.address)->result = ex(it+(j->first).lag-1,s.nbr);
	  break;
	case 1:
	  //	  ((tm_it)s.address)->result = y(it+(j->first).lag-1,s.nbr);
	  (s.address)->result = y(it+(j->first).lag-1,s.nbr);
// 	  std::cout << s.nbr << " " << " " << (j->first).name << " " << it+(j->first).lag-1 << " " << y(it+(j->first).lag-1,s.nbr) << "\n";
	  break;
	case 3:
	  //	  ((tm_it)s.address)->result = param[s.nbr];
	  (s.address)->result = param[s.nbr];
	  break;
	default:
	  std::cerr << "execute_tree init: unknown variable type\n";
	  exit(-1);
	}
    }

  g_v = vec_double();
  g_i = vec_int();
  g_r = vec_int();
  int old_order = 0;

  for (tm_it i = model.begin(); i != model.end(); i++)
    {
      if (i->code == 0)
	{
	  continue;
	}

      double a1, a2;
      if ( i->arg1 == NODE0 )
	{
	  a1 = 0.0;
	}
      else if ( i->arg1 == NODE1)
	{
	  a1 = 1.0;
	}
      else
	{
	  //	  a1 = ((tm_it)i->arg1)->result;
	  a1 = (i->arg1)->result;
	}

      if ( i->arg2 == NODE0 )
	{
	  a2 = 0.0;
	}
      else if ( i->arg2 == NODE1)
	{
	  a2 = 1.0;
	}
      else
	{
	  //	  a2 = ((tm_it)i->arg2)->result;
	  a2 = (i->arg2)->result;
	}
      
      switch ((*i).code)
	{
	case PLUS:
	  (*i).result=a1+a2;
	  break;
	case MINUS:
	  if ((*i).arg1 != NODE0)
	    {
	  (*i).result=a1-a2;
	    }
	  else
	    {
	  (*i).result=-a2;
	    }
	  break;
	case DIVIDE:
	  (*i).result=a1/a2;
	  break;
	case TIMES:
	  (*i).result=a1*a2;
	  break;
	case EXP:
	  (*i).result=exp(a2);
	  break;
	case LOG:
	  (*i).result=log(a2);
	  break;
	case LOG10:
	  (*i).result=log10(a2);
	  break;
	case COS:
	  (*i).result=cos(a2);
	  break;
	case SIN:
	  (*i).result=sin(a2);
	  break;
	case TAN:
	  (*i).result=tan(a2);
	  break;
	case ACOS:
	  (*i).result=acos(a2);
	  break;
	case ASIN:
	  (*i).result=asin(a2);
	  break;
	case ATAN:
	  (*i).result=atan(a2);
	  break;
	case COSH:
	  (*i).result=cosh(a2);
	  break;
	case SINH:
	  (*i).result=sinh(a2);
	  break;
	case TANH:
	  (*i).result=tanh(a2);
	  break;
	case ACOSH:
#ifndef __MINGW32__
	  (*i).result=acosh(a2);
#else
	  (*i).result=0;
#endif
	  break;
	case ASINH:
#ifndef __MINGW32__
	  (*i).result=asinh(a2);
#else
	  (*i).result=0;
#endif
	  break;
	case ATANH:
#ifndef __MINGW32__
	  (*i).result=atanh(a2);
#else
	  (*i).result=0;
#endif
	  break;
	case SQRT:
	  (*i).result=sqrt(a2);
	  break;
	case POWER:
	  if (a2 == 0)
	    {
	      (*i).result = 1;
	      break;
	    }
	  if (a1 == 0)
	    {
	      (*i).result = 0;
	    }
	  else
	    {
	      (*i).result=pow(a1,a2);
	    }
	  break;
	case EQUAL_SIGN:
	  if (i->no == 0)
	    {
	      zz[i->ref]=a2;
	    }
	  else
	    {
			t_h_deriv::const_iterator it = deriv_col.find(i->no);
			if (it != deriv_col.end()) {
				if ((*it).second.deriv_order > old_order)
				{
					if (flag) {
					md_st = new SparseTensor((*it).second.deriv_order+1,sm.dyn_endo_nbr+sm.exo_nbr,sm.endo_nbr);
					md.insert(md_st);
					}
					make_sparse(g_v, g_i, g_r, old_order);
					old_order = (*it).second.deriv_order;
					g_v = vec_double();
					g_i = vec_int();
					g_r = vec_int();
				}
				int jj = (*it).second.col;
				g_v.push_back(a2);
				g_i.push_back(jj);
				g_r.push_back(i->ref+1);
				if (flag) {
					IntSequence ss((*it).second.intseq);
					ss.sort();
					(*md_st).insert(ss,i->ref,a2);
				}
			}
		}
	  break;
	default:
	  std::cerr << "Missing code:" << i->code;
	}
//        std::cout << (*i).code << " " << a1 << " " << a2 << " " << (*i).result << "\n";;
//         if ((*i).code == EQUAL_SIGN)
//   	{
//   	  std::cout << "Eq " << i->ref << "\n";
//   	}
      if ( i == end_node )
	{
	  break;
	}
    }
  if (flag == 1)
    {
      make_sparse(g_v, g_i, g_r, old_order);
    }
}

void make_sparse(vec_double &g_v, vec_int &g_c, vec_int &g_r, int deriv_order)
{
  unsigned int n = g_v.size();
  typedef mtl::dense1D<vec_int> vec_of_vec;
  vec_of_vec h(sm.endo_nbr);
  for ( unsigned int i = 0; i < n; i++)
    {
      int k = g_r[i]-1;
      if ( h[k].size() == 0)
	{
	  h[k] = vec_int(1);
	  h[k][0] = i;
	}
      else
	{
	  h[k].push_back(i);
	}
    }
  int * ptr = new int[sm.endo_nbr+1];
  int * pptr = ptr;
  int * ind = new int[n];
  int * pind = ind;
  double * val = new double[n];
  double * pval = val;
  
  int count = 1;
  for (vec_of_vec::iterator i = h.begin(); i != h.end(); i++)
    {
      *(pptr++) = count;
      for (vec_int::iterator j = (*i).begin(); j != (*i).end(); j++)
	{
	  *(pval++) = g_v[*j];
	  *(pind++) = g_c[*j];
	  ++count;
	}
    }
  *pptr = count;
  f[deriv_order+1] = Matx(sm.endo_nbr,FFSTensor::calcMaxOffset(sm.dyn_endo_nbr+sm.exo_nbr, deriv_order),n,
			  val,ptr,ind);
}


