#include <mtl/mtl.h>
#include <mtl/lu.h>

template<class MatrixA, class VectorB, class MatrixC, class MatrixD> 
inline void lu_solve_mat(const MatrixA &lu, const VectorB &p, const MatrixC &B, MatrixD &X)
{
  vec_double temp(lu.nrows());
  unsigned int n = B.ncols();
  for(unsigned int i = 0; i < n; i++)
    {
      mtl::lu_solve(lu, p, columns(B)[i],temp);
      copy(temp,columns(X)[i]);
    }
}

template<class MatrixA, class VectorB, class MatrixC, class MatrixD> 
inline void lu_solve_mat(const MatrixA &lu, const VectorB &p, const MatrixC &B, MatrixD &X, int info)
{
  vec_double temp(lu.nrows());
  unsigned int n = B.ncols();
  for(unsigned int i = 0; i < n; i++)
    {
      lu_solve(lu, p, columns(B)[i],temp,info);
      copy(temp,columns(X)[i]);
    }
}

template <class DenseMatrix, class VectorB, class VectorX, class Pvector>
void
lu_solve(const DenseMatrix &LU, const Pvector& pvector, 
	 const VectorB &b, VectorX &x, int info)
{
  typedef typename Pvector::size_type p_int;

  copy(b, x);

  /* use the permutation vector to modify the starting vector 
   *  to account for the permutations in LU
   */
  for(p_int i=0; i < (p_int)info; i++) {
    p_int perm = pvector[i]-1;       // permutations stored in 1's offset
    if(i != perm)
      std::swap(x[i], x[perm]);
  }
  
  /* solve  Ax = b  ->  LUx = b  ->  Ux = L^-1 b 
   *  which we solve in two steps
   *  1) y = L^-1 b
   *  2) x = U^-1 y
   */
  typename mtl::triangle_view<DenseMatrix, mtl::unit_lower>::type L(LU);
  typename mtl::triangle_view<DenseMatrix, mtl::upper>::type U(LU);

  tri_solve(L, x);
  for (unsigned int i=info; i< x.size();  ++i)
    {
      if (fabs(x[i]) > 1e-6)
	{
	  std::cout << "Incompatible singular system in lu_solve\n";
	  std::cout << x[i] << "\n";
	  exit(1);
	}
    }
  typedef typename DenseMatrix::submatrix_type DenseSubMatrix; 
  DenseSubMatrix  LU1=LU.sub_matrix(0,info-1,0,info-1);
  typename mtl::triangle_view<DenseSubMatrix, mtl::upper>::type U1(LU1);
  tri_solve(U1, x(0,info-1));
}


