#include <mtl/mtl.h>
#include <mtl/lu.h>
#include "my_mtl.h"

typedef MatA MatrixA;
typedef vec_int VectorB;
typedef MatA MatrixC;
typedef MatA MatrixD;

typedef MatA DenseMatrix;
typedef vec_double VectorBB;
typedef vec_double VectorX;
typedef vec_int Pvector;

void
lu_solve(const DenseMatrix &, const Pvector&, 
	 const VectorBB &, VectorX &, int );

void lu_solve_mat1(MatrixA &lu, VectorB &p, MatrixC &B, MatrixD &X, int info)
{
  vec_double temp(lu.nrows());
  int n = B.ncols();
  for(unsigned int i = 0; i < n; i++)
    {
      vec_double temp1(B.nrows());
      copy(columns(B)[i],temp1);
      lu_solve(lu, p, temp1,temp,info);
      copy(temp,columns(X)[i]);
    }
}

void
lu_solve(const DenseMatrix &LU, const Pvector & pvector, 
	 const VectorBB &b, VectorX &x, int info)
{
  typedef  Pvector::size_type p_int;

  copy(b, x);

  /* use the permutation vector to modify the starting vector 
   *  to account for the permutations in LU
   */
  print_vector(pvector);
  for(p_int i=0; i < info; i++) {
    p_int perm = pvector[i]-1;       // permutations stored in 1's offset
    if(i != perm)
        std::swap(x[i], x[perm]);
  }
  
  /* solve  Ax = b  ->  LUx = b  ->  Ux = L^-1 b 
   *  which we solve in two steps
   *  1) y = L^-1 b
   *  2) x = U^-1 y
   */
   triangle_view<DenseMatrix, unit_lower>::type L(LU);
   triangle_view<DenseMatrix, upper>::type U(LU);

  tri_solve(L, x);
  for (int i=info; i< x.size();  ++i)
    {
      if (fabs(x[i]) > 1e-12)
	{
	  std::cout << "Incompatible singular system in lu_solve\n";
	  std::cout << x[i] << "\n";
	  exit(1);
	}
    }
  typedef  DenseMatrix::submatrix_type DenseSubMatrix; 
  DenseSubMatrix  LU1=LU.sub_matrix(0,info-1,0,info-1);
   triangle_view<DenseSubMatrix, upper>::type U1(LU1);
  tri_solve(U1, x(0,info-1));
  print_vector(x);
}


