#ifndef MJDGEES_H
#define MJDGEES_H

typedef int logical;
typedef logical (*L_fp)(...);

#include <complex>

#if defined MTL_FCALL
#undef MTL_FCALL
#define MTL_FCALL(x) x##_
#endif

namespace mtl_lapack_dispatch {
  extern "C" {
  void MTL_FCALL(dgees)(const char &jobvs, const char &sort, L_fp select, 
			const int & n, double a[], const int & lda, int & sdim,
			double wr[], double wi[], double vs[], 
			const int & ldvs, double work[], const int & lwork, 
			int bwork[], int & info );
  }

  inline void gees(const char &jobvs, const char &sort, L_fp select, 
		    const int & n, double a[], const int & lda, int & sdim,
		    std::complex<double> w[], double vs[], 
		    const int & ldvs, int & info )
    {
      double *wr = new double[n];
      double *wi = new double[n];
      int lwork = 10*n;
      double * work = new double[lwork];
      int * bwork = new int[n];

      MTL_FCALL(dgees)(jobvs, sort, select, n, a, lda, sdim, wr, wi, vs, ldvs,
		       work, lwork, bwork, info);
      for(int i=0; i<n; i++)
	{
	  w[i] = std::complex<double>(wr[i],wi[i]);
	}
      delete [] wr;
      delete [] wi;
      delete [] work;
      delete [] bwork;
    }
}

namespace mtl2lapack {
  template <class LapackMatA, class LapackMatVS, class VectorComplex>
  int gees(int jobv,
	   LapackMatA& a,        /* N x N */
	   VectorComplex& alpha,     /* N x 1 */
	   LapackMatVS& vs,  /* LDVS x N */
	   int & sdim,
	   L_fp delctg)
  {
    char _jobvs, _sort;

    int  _info;
 
    /*determine _jobvsr and _jobvsl */
    switch (jobv) {
    case 0: /* GEES_CALC_NONE_NOSORT: */
      _jobvs = 'N';
      _sort = 'N';
      break;
    case 1: /* GEES_CALC_SCHUR_NOSORT: */
      _jobvs = 'V';
      _sort = 'N';
      break;
    case 2:  /* GEES_CALC_NONE_SORT: */
      _jobvs = 'N';
      _sort = 'S';
      break;
    case 3:  /* GEES_CALC_SCHUR_SORT:*/
      _jobvs = 'V';
      _sort = 'S';
      break;
    default:
      assert(0);
    }

    int _lda = a.minor();
    int a_n = a.major();
    int _ldvs = vs.minor();
  
    /* make sure matrix is square */
    if (a.nrows() != a.ncols())
      return -100;

    /* make sure result eigenvalue vector is long enough */
    if (int(alpha.size()) != a_n)
      return -101;

    if (_jobvs == 'V')
      if (int(vs.nrows()) < a_n || int(vs.ncols()) < a_n)
        return -102;

    /* call dispatcher */
    mtl_lapack_dispatch::gees(_jobvs, _sort, delctg, a_n, a.data(), 
			      _lda, sdim, alpha.data(), vs.data(), _ldvs, 
			      _info);

    return _info;

  }
  enum GEES_JOBV {GEES_CALC_NONE_NOSORT, GEES_CALC_SCHUR_NOSORT, 
		  GEES_CALC_NONE_SORT, GEES_CALC_SCHUR_SORT};

}
#endif







