#if !defined MJDGGES_H
#define MJDGGES_H

typedef int logical;
typedef logical (*L_fp)(...);

#include <complex>

#if defined MTL_FCALL
#undef MTL_FCALL
#define MTL_FCALL(x) x##_
#endif

logical my_criteria(double *alphar, double *alphai, double *beta); // defined in utils.cc

namespace mtl_lapack_dispatch {
  extern "C" {
  void MTL_FCALL(dgges)(const char &jobvsl, const char &jobvsr, const char &sort, 
	      L_fp delctg, const int & n, double da[], const int & lda, 
	      double db[], const int & ldb, int & sdim, double alphar[], 
	      double alphai[], double beta[], double vsl[], const int & ldvsl,
	      double vsr[], const int & ldvsr, double dwork[], 
	      const int & ldwork, int bwork[], int & info);
  }

  inline void gges(const char & jobvsl, const char & jobvsr, const char & sort,
		   L_fp delctg, const int & n, double da[], const int &lda, 
		   double db[], const int & ldb, int & sdim, 
		   std::complex<double> alpha[], double beta[], double vsl[], 
		   const int & ldvsl, double vsr[], const int & ldvsr, 
		   int & info)
    {
      double * alphar = new double[n];
      double * alphai = new double[n];
      int ldwork = 7*(n+1)+16;
      double * dwork = new double[ldwork];
      int * bwork = new int[n];

      MTL_FCALL(dgges)(jobvsl, jobvsr, sort, delctg, n, da, lda, db,
		       ldb, sdim, alphar, alphai, beta, vsl, ldvsl, vsr,
		       ldvsr, dwork, ldwork, bwork, info);
      for(int i=0; i<n; i++)
	{
	  alpha[i] = std::complex<double>(alphar[i],alphai[i]);
	}
      delete [] alphar;
      delete [] alphai;
      delete [] dwork;
      delete [] bwork;
    }
}

namespace mtl2lapack {
  template <class LapackMatA, class LapackMatB, class LapackMatVSL, 
            class LapackMatVSR, class VectorComplex, class VectorReal>
  int gges(int jobv,
	   LapackMatA& a,        /* N x N */
	   LapackMatB& b,        /* N x N */
	   VectorComplex& alpha,     /* N x 1 */
	   VectorReal& beta,       /* N x 1 */
	   LapackMatVSL& vsl,  /* LDVSL x N */
	   LapackMatVSR& vsr,  /* LDVSR x N */
	   int & sdim,
	   L_fp delctg)
  {
    char _jobvsr, _jobvsl, _sort;

    int  _info;
 
    /*determine _jobvsr and _jobvsl */
    switch (jobv) {
    case 0: /* GGES_CALC_LEFT_NOSORT: */
      _jobvsr = 'N';
      _jobvsl = 'V';
      _sort = 'N';
      break;
    case 1: /* GGES_CALC_RIGHT_NOSORT: */
      _jobvsr = 'V';
      _jobvsl = 'N';
      _sort = 'N';
      break;
    case 2:  /* GGES_CALC_BOTH_NOSORT: */
      _jobvsr = 'V';
      _jobvsl = 'V';
      _sort = 'N';
      break;
    case 3:  /* GGES_CALC_NONE_NOSORT:*/
      _jobvsr = 'N';
      _jobvsl = 'N';
      _sort = 'N';
      break;
    case 4: /* GGES_CALC_LEFT_SORT: */
      _jobvsr = 'N';
      _jobvsl = 'V';
      _sort = 'S';
      break;
    case 5: /* GGES_CALC_RIGHT_SORT: */
      _jobvsr = 'V';
      _jobvsl = 'N';
      _sort = 'S';
      break;
    case 6:  /* GGES_CALC_BOTH_SORT: */
      _jobvsr = 'V';
      _jobvsl = 'V';
      _sort = 'S';
      break;
    case 7:  /* GGES_CALC_NONE_SORT:*/
      _jobvsr = 'N';
      _jobvsl = 'N';
      _sort = 'S';
      break;
    default:
      assert(0);
    }

    int             _lda = a.minor();
    int             _ldb = b.minor();
    int             a_n = a.major();
    int             _ldvsl = vsl.minor();
    int             _ldvsr = vsr.minor();
  
    /* make sure matrix is square */
    if (a.nrows() != a.ncols())
      return -100;

    if (b.nrows() != b.ncols())
      return -101;

    /* make sure matrix a and d have same dimension */
    if (a.nrows() != b.nrows())
      return -102;

    /* make sure result eigenvalue vector is long enough */
    if (int(alpha.size()) != a_n)
      return -104;

    if (int(beta.size()) != a_n)
      return -104;

    if (_jobvsl == 'V')
      if (int(vsl.nrows()) < a_n || int(vsl.ncols()) < a_n)
        return -105;

    if (_jobvsr == 'V')
      if (int(vsr.nrows()) < a_n || int(vsr.ncols()) < a_n )
	return -106;

    /* call dispatcher */
    mtl_lapack_dispatch::gges(_jobvsl, _jobvsr, _sort, delctg, a_n, a.data(), 
			      _lda, b.data(), _ldb, sdim, alpha.data(),
			      beta.data(), vsl.data(), _ldvsl, vsr.data(), 
			      _ldvsr, _info);

    return _info;

  }
  enum GGES_JOBV {GGES_CALC_LEFT_NOSORT, GGES_CALC_RIGHT_NOSORT, 
		  GGES_CALC_BOTH_NOSORT, GGES_CALC_NONE_NOSORT, 
		  GGES_CALC_LEFT_SORT, GGES_CALC_RIGHT_SORT, 
		  GGES_CALC_BOTH_SORT, GGES_CALC_NONE_SORT};

}
#endif







