#include "dynare2.h"
#include <iostream>
#include <string>
#include <sstream>
#include "mtl/mtl2lapack.h"
#include "mjdpotrf.h"
#include "moments.h"

extern type_symbol_table symbols;
extern vec_double param;
typedef mtl2lapack::lapack_matrix<double>::type Lapack_Mat;
Moments moments;

// get_param shouldn't be in this file
double get_param(const char *name)
{
  type_symbol_table::iterator i = symbols.find(name);
  if (i == symbols.end())
    {
      std::cout << "Unknown parameter: " << name;
      exit(1);
    }
  if ((i->second).type != 3)
    {
      std::cout << "Non-parameter used in VCOV expression: " << name;
      exit(1);
    }

  return param[(i->second).nbr];
}

void Moments::push(const char *number)
{
//  std::cout << *number << *(number+1) << "\n"; 
  if (vcov_col > var_nbr)
    {
      std::cout << "Too many columns in VCOV\n";
      exit(1);
    }
  if (vcov_row > var_nbr)
    {
      std::cout << "Too many rows in VCOV\n";
      exit(1);
    }
  std::istringstream is(number);
  double dvalue;
  is >> dvalue;
  if (vcov_triang == 0 && vcov_row > vcov_col && dvalue != vcov(vcov_col,vcov_row))
    {
      std::cout << "VCOV isn't symmetrical\n";
    }
  vcov(vcov_row,vcov_col) = dvalue;
  ++vcov_col;
}

void Moments::push(const char *number, const int sign)
{
//  std::cout << *number << *(number+1) << sign << "\n"; 
  if (vcov_col > var_nbr)
    {
      std::cout << "Too many columns in VCOV\n";
      exit(1);
    }
  if (vcov_row > var_nbr)
    {
      std::cout << "Too many rows in VCOV\n";
      exit(1);
    }
  std::istringstream is(number);
  double dvalue;
  is >> dvalue;
  dvalue = sign * dvalue;
  if (vcov_triang == 0 && vcov_row > vcov_col && dvalue != vcov(vcov_col,vcov_row))
    {
      std::cout << "VCOV isn't symmetrical\n";
    }
  vcov(vcov_row,vcov_col) = dvalue;
  ++vcov_col;
}

void Moments::push(const double dvalue)
{
  if (vcov_col > var_nbr)
    {
      std::cout << "Too many columns in VCOV\n";
      exit(1);
    }
  if (vcov_row > var_nbr)
    {
      std::cout << "Too many rows in VCOV\n";
      exit(1);
    }
  if (vcov_triang == 0 && vcov_row > vcov_col && dvalue != vcov(vcov_col,vcov_row))
    {
      std::cout << "VCOV isn't symmetrical\n";
    }
  vcov(vcov_row,vcov_col) = dvalue;
  ++vcov_col;
}

void Moments::row_inc()
{
  ++vcov_row;
  if (vcov_row == 1 && vcov_col == 0)
    {
      vcov_triang = 1;
    }
  if (vcov_triang == 0)
    {
      if (vcov_col < var_nbr)
	{
	  std::cout << "Incomplete row in VCOV\n";
	  exit(1);
	}
      else if (vcov_col > var_nbr)
	{
	  std::cout << "Too many elements in a  row of VCOV\n";
	  exit(1);
	}
    }
  else
    {
      if (vcov_col < vcov_row-1)
	{
	  std::cout << "Incomplete row in triangular VCOV\n";
	  exit(1);
	}
      else if (vcov_col >= vcov_row)
	{
	  std::cout << "Too many elements in a  row of triangular VCOV\n";
	  exit(1);
	}
    }
  vcov_col = 0;
}

void Moments::set_variance(const char* name, const double dvalue)
{
  int k = symbols[name].nbr;
  vcov(k,k) = dvalue;
}    
