#include "dynare2.h"
#include <mtl/lu.h>

int lnsrch(vec_double &, double, vec_double &, vec_double &, double, vec_double &);

extern struct model_struct sm;
extern tm_it end_resid;
extern vec_double ys, param;
extern MatA y;
extern TensorContainer<SparseTensor> md;
double tolx = 3.7e-11;


void nlsolve(vec_double & x)
{
  
  vec_double fvec(sm.endo_nbr), p(sm.endo_nbr), junk1;
  vec_int ptr(sm.endo_nbr), junk2, junk3;

  y = MatA(-sm.maxlag+sm.maxlead+1,sm.endo_nbr);

  for(int i = 0; i < -sm.maxlag+sm.maxlead+1; i++)
    {
      copy(x, rows(y)[i]);
    }
  print_all_matrix(y);
  double stpmax = 100*sqrt(std::max(dot(x,x),(double)x.size()));
  int it = 1-sm.maxlag;
  execute_tree(md, fvec, junk1, junk2, junk3, end_resid, it, 0);
  std::cout << "fvec\n";
  for (unsigned int i = 1; i <= fvec.size(); ++i)
    {
      std::cout << i << " " << fvec[i-1] << "\n";
    }

  double f = 0.5*sum_squares(fvec);
  if (isnan(f))
    {
      std::cout << "f is NaN\n";
      for (unsigned int i = 1; i <= fvec.size(); ++i)
	{
	  std::cout << i << " " << fvec[i] << "\n";
	}
      exit(1);
    }

  while (1)
    {
      double err = fabs(fvec[mtl::max_abs_index(fvec)]);
      std::cout << "Err: " << err << "\n";
      if (err < 1e-10)
	{
	  break;
	}

      for(int i = 0; i < -sm.maxlag+sm.maxlead+1; i++)
	{
	  copy(x, rows(y)[i]);
	}
      MatA Jacobian(sm.endo_nbr,sm.endo_nbr);
      jacobian(fvec, Jacobian);
//       std::cout << "Jacobian\n";
//       print_all_matrix(Jacobian);
      vec_double g(sm.endo_nbr);
      mult(trans(Jacobian),fvec,g);
      int info=mtl::lu_factor(Jacobian, ptr);
      if (info != 0)
	{
	  std::cout << "lu_factor info" << info << "\n";
	  exit(-1);
	}
      print_vector(ptr);
      mtl::lu_solve(Jacobian, ptr,fvec, p);
      scale(p,-1.0);
      lnsrch(x, f, g, p, stpmax, fvec);
    }
}

int lnsrch(vec_double & x, double f, vec_double & g, vec_double & p, double stpmax, vec_double & fvec)
{
  double alf=1e-4;
  double fold = f;

  double sum = sqrt(sum_squares(p));
  if ( sum > stpmax)
    {
      scale(p, stpmax/sum);
    }
  
  double slope=dot(g,p);


  double test = 0;
  vec_double::iterator j=x.begin();
  for (vec_double::iterator i = p.begin(); i != p.end(); ++i)
    {
      double tmp = fabs(*i) / ((fabs(*j) > 1) ? *j : 1); 
      if ( tmp > test )
	{
	  test = tmp;
	}
      ++j;
    }
  double alamin = tolx / test;

  double alam = 1;
  while (1)
    {
      if (alam < alamin)
	{
	  return 1;
	}
      add(scaled(p,alam),x);
      for(int i = 0; i < -sm.maxlag+sm.maxlead+1; i++)
	{
	  copy(x,rows(y)[i]);
	}
      vec_double junk1;
      vec_int junk2, junk3;
      execute_tree(md, fvec, junk1, junk2, junk3, end_resid, 1-sm.maxlag, 0);
      double f =0.5*dot(fvec,fvec);
      if (isnan(f))
	{
	  std::cerr << "f is NaN\n";
	  for (unsigned int i = 1; i <= fvec.size(); ++i)
	    {
	      std::cout << i << " " << fvec[i] << "\n";
	    }
	  exit(1);
	}
      if (f <= fold+alf*alam*slope)
	{
	  std::cout << "Alam " << alam << "\n";
	  return 0;
	}
      double tmplam, f2, fold2, alam2;
      if (alam == 1)
	{
	  tmplam = -slope/(2*(f-fold-slope));
	}
      else
	{
	  double rhs1 = f-fold-alam*slope;
	  double rhs2 = f2-fold2-alam2*slope;
	  double a = (rhs1/(alam*alam)-rhs2/(alam2*alam2))/(alam-alam2);
	  double b=(-alam2*rhs1/(alam*alam)+alam*rhs2/(alam2*alam2))/(alam-alam2);
	  if (a == 0)
	    {
	      tmplam = -slope/(2*b);
	    }
	  else
	    {
	      double disc = b*b-3*a*slope;
	      if (disc < 0)
		{
		  std::cerr << "ERROR: Roundoff problem in nlsearch\n";
		  exit(1);
		}
	      else
		{
		  tmplam=(-b+sqrt(disc))/(3*a);
		}
	    }
	  if (tmplam > 0.5*alam)
	    {
	      tmplam = 0.5*alam;
	    }
	}
      alam2=alam;
      f2=f;
      fold2=fold;
      alam=std::max(tmplam,0.1*alam);
    }
}
