#include "dynare2.h"

extern type_variable_table variables;
extern type_symbol_table symbols;
extern struct model_struct sm;
extern  t_var_deriv_index var_deriv_index;
extern mtl::dense1D<s_mixed> mixed;
extern vec_double ys, exe;
void state(t_state &);

struct variable_comp1 : public std::binary_function <type_variable_table::iterator, type_variable_table::iterator, bool> 
{
  inline bool operator()(const type_variable_table::iterator & i1, const type_variable_table::iterator & i2)
  {
    if ( (i1->second).type > (i2->second).type )
      {
	return 1;
      }
    else if ( (i1->second).type < (i2->second).type )
      {
	return 0;
      }
    else if ( (i1->first).lag < (i2->first).lag)
      {
	return 0;
      }
    else if ( (i1->first).lag > (i2->first).lag)
      {
	return 1;
      }
    else
      {
	return ( (i1->second).nbr < (i2->second).nbr );
      }
  }
};


struct v_cmp1 : public std::binary_function <s_mixed, s_mixed, bool> 
{
  inline bool operator()(const s_mixed & i1, const s_mixed & i2)
  {
    if (i1.mix < i2.mix)
      {
	return 1;
      }
    else if (i1.mix > i2.mix)
      {
	return 0;
      }
    else
      {
	return (i1.id < i2.id);
      }
  }
};

struct v_cmp2 : public std::binary_function <s_mixed, s_mixed, bool> 
{
  inline bool operator()(const s_mixed & i1, const s_mixed & i2)
  {
    return (i1.id < i2.id);
  }
};

void sort_deriv_variables( void )
{

  mixed.resize(sm.endo_nbr);
  MatA laglead(sm.endo_nbr,2);

  for (type_variable_table::iterator i=variables.begin(); i != variables.end(); i++)
    {
      if ((i->second).type == 1)
	{
	  int nbr = (i->second).nbr;
	  mixed[nbr].id = nbr;
	  if ( (i->first).lag < laglead(nbr,0) )
	    {
	      laglead(nbr,0) = (i->first).lag;
	    }
	  else if ( (i->first).lag > laglead(nbr,1) )
	    {
	      laglead(nbr,1) = (i->first).lag;
	    }
	  if ( (i->first).lag < 0 )
	    {
	      if ( mixed[nbr].mix == 1)
		{
		  mixed[nbr].mix = 2;
		}
	      else if ( mixed[nbr].mix > 2)
		{
		  mixed[nbr].mix = 3;
		}
	      else
		{
		  mixed[nbr].mix = 2;
		}
	      ++sm.lagged_nbr;
	    }
	  else if ( (i->first).lag > 0 )
	    {
	      if ( mixed[nbr].mix == 0 || mixed[nbr].mix == 4 )
		{
		  mixed[nbr].mix = 4;
		}
	      else if ( mixed[nbr].mix == 1 )
		{
		  mixed[nbr].mix = 4;
		}
	      else
		{
		  mixed[nbr].mix = 3;
		}
	      ++sm.fwrd_nbr;
	    }
	  else if (mixed[nbr].mix == 0)
	    {
	      mixed[nbr].mix = 1;
	    }
	}
    }
  
  for(int i = sm.maxlag; i < 0; ++i)
    {
      for (type_symbol_table::iterator j=symbols.begin(); j != symbols.end(); j++)
	{
	  if ( (j->second).type == 1 )
	    {
	      if ( laglead((j->second).nbr,0) < i )
		{
		  s_variable_key sk;
		  sk.name = j->first;
		  sk.lag = i;
		  type_variable_table::iterator k = variables.find(sk);
		  if ( k == variables.end() )
		    {
		      add_variable(j->first,sk.lag,1);
		    }
		}
	    }
	}
    }

  for(int i = 1; i <= sm.maxlead; ++i)
    {
      for (type_symbol_table::iterator j=symbols.begin(); j != symbols.end(); j++)
	{
	  if ( (j->second).type == 1 )
	    {
	      if ( laglead((j->second).nbr,1) > i )
		{
		  s_variable_key sk;
		  sk.name = j->first;
		  sk.lag = i;
		  type_variable_table::iterator k = variables.find(sk);
		  if ( k == variables.end() )
		    {
		      add_variable(j->first,sk.lag,1);
		    }
		}
	    }
	}
    }

  sort(mixed.begin(),mixed.end(),v_cmp1());
  int j = 0;
  for(mtl::dense1D<s_mixed>::iterator i=mixed.begin(); i != mixed.end(); i++)
    {
      (*i).nbr = j++;
      switch((*i).mix)
	{
	case 1:
	  ++sm.static_nbr;
	  break;
	case 2:
	  ++sm.c_lagged_nbr;
	  break;
	case 3:
	  ++sm.c_lagged_nbr;
	  ++sm.c_fwrd_nbr;
	  ++sm.c_both_nbr;
	  break;
	case 4:
	  ++sm.c_fwrd_nbr;
	  break;
	default:
	  std::cerr << "Error in sort_deriv_var: mix can't be 0\n";
	}
    }

  sort(mixed.begin(),mixed.end(),v_cmp2());

  vec_double tempy(ys.size());

  for (type_variable_table::iterator i=variables.begin(); i != variables.end(); i++)
    {
      if ( ( (*i).second).type == 1)
  	{
	  int k = ( (*i).second).nbr;
	  int k1 = mixed[k].nbr; 
  	  ((*i).second).nbr = k1;
	  tempy[k1] = ys[k];
  	}
    }
  
  copy(tempy,ys);

  for (type_variable_table::iterator i=variables.begin(); i != variables.end(); i++)
    {
      if ( ((*i).second).type <= 1)
	{
	  var_deriv_index.push_back(i);
	}
    }

  sort(var_deriv_index.begin(),var_deriv_index.end(),variable_comp1());

  state(sm.state_vec);
  // pointer to first static variables in z vector 
  sm.p_static = sm.fwrd_nbr;
  // pointer to current value of first foward-looking variable in z vector
  sm.p_c_lagged = sm.p_static+sm.static_nbr;
  // pointer to current value of first predetermined variable in z vector
  sm.p_c_fwrd = sm.p_c_lagged+sm.c_lagged_nbr-sm.c_both_nbr;
  // pointer to first futur value of forward-looking variable
  sm.p_fwrd = 0;
  // pointer to first previous value of lagged variable
  sm.p_lagged = sm.fwrd_nbr+sm.endo_nbr;
  // pointer to first exogenous variable
  sm.p_exo = sm.fwrd_nbr+sm.endo_nbr+sm.s_lagged_nbr;

}
