#include "dynare2.h"
#include <strstream>
#include "mtl/mtl2lapack.h"

extern type_symbol_table symbols;
extern vec_double param;
extern model_struct sm;
typedef mtl2lapack::lapack_matrix<double>::type Lapack_Mat;
extern Lapack_Mat vcov;
int vcov_row = 0;
int vcov_col = 0;
int vcov_triang = 0;

double get_param(const char *name)
{
  type_symbol_table::iterator i = symbols.find(name);
  if (i == symbols.end())
    {
      cerr << "Unknown parameter: " << name;
      exit(1);
    }
  if ((i->second).type != 3)
    {
      cerr << "Non-parameter used in vcov expression: " << name;
      exit(1);
    }

  return param[(i->second).nbr];
}

void vcov_push(const char *number)
{
  if (vcov_col > sm.exo_nbr)
    {
      cerr << "Too many columns in VCOV\n";
      exit(1);
    }
  if (vcov_row > sm.exo_nbr)
    {
      cerr << "Too many rows in VCOV\n";
      exit(1);
    }
  istrstream is(number);
  double dvalue;
  is >> dvalue;
  if (vcov_triang == 0 && vcov_row > vcov_col && dvalue != vcov(vcov_col,vcov_row))
    {
      cerr << "VCOV isn't symmetrical\n";
    }
  vcov(vcov_row,vcov_col) = dvalue;
  ++vcov_col;
}

void vcov_push(const double dvalue)
{
  if (vcov_col > sm.exo_nbr)
    {
      cerr << "Too many columns in VCOV\n";
      exit(1);
    }
  if (vcov_row > sm.exo_nbr)
    {
      cerr << "Too many rows in VCOV\n";
      exit(1);
    }
  if (vcov_triang == 0 && vcov_row > vcov_col && dvalue != vcov(vcov_col,vcov_row))
    {
      cerr << "VCOV isn't symmetrical\n";
    }
  vcov(vcov_row,vcov_col) = dvalue;
  ++vcov_col;
}

void vcov_row_inc()
{
  ++vcov_row;
  if (vcov_row == 1 && vcov_col == 0)
    {
      vcov_triang = 1;
    }
  if (vcov_triang == 0)
    {
      if (vcov_col < sm.exo_nbr)
	{
	  cerr << "Incomplete row in VCOV\n";
	  exit(1);
	}
      else if (vcov_col > sm.exo_nbr)
	{
	  cerr << "Too many elements in a  row of VCOV\n";
	  exit(1);
	}
    }
  else
    {
      if (vcov_col < vcov_row-1)
	{
	  cerr << "Incomplete row in triangular VCOV\n";
	  exit(1);
	}
      else if (vcov_col >= vcov_row)
	{
	  cerr << "Too many elements in a  row of triangular VCOV\n";
	  exit(1);
	}
    }
  vcov_col = 0;
}
    

