/* $Header: /var/lib/cvs/dynare_cpp/sylv/cc/GeneralMatrix.h,v 1.3 2004/11/24 20:41:59 kamenik Exp $ */

/* Tag $Name:  $ */

#ifndef GENERAL_MATRIX_H
#define GENERAL_MATRIX_H

#include "Vector.h"

class GeneralMatrix;

class ConstGeneralMatrix {
	friend class GeneralMatrix;
protected:
	ConstVector data;
	int rows;
	int cols;
	int ld;
public:
	ConstGeneralMatrix(const double* d, int m, int n)
		: data(d, m*n), rows(m), cols(n), ld(m) {}
	ConstGeneralMatrix(const GeneralMatrix& m);
	ConstGeneralMatrix(const GeneralMatrix& m, int i, int j, int nrows, int ncols);
	ConstGeneralMatrix(const ConstGeneralMatrix& m, int i, int j, int nrows, int ncols);
	virtual ~ConstGeneralMatrix() {}

	const double& get(int i, int j) const
		{return data[j*ld+i];}
	int numRows() const {return rows;}
	int numCols() const {return cols;}
	int getLD() const {return ld;}
	const double* base() const {return data.base();}
	const ConstVector& getData() const {return data;}

	double getNormInf() const;
	double getNorm1() const;
	/* x = x + this*d */
	void multaVec(Vector& x, const ConstVector& d) const;
	/* x = x + this'*d */
	void multaVecTrans(Vector& x, const ConstVector& d) const;
	/* x = x - this*d */
	void multsVec(Vector& x, const ConstVector& d) const;
	/* x = x - this'*d */
	void multsVecTrans(Vector& x, const ConstVector& d) const;
	/* m = inv(this)*m */
	void multInvLeft(GeneralMatrix& m) const
		{multInvLeft("N", m);}
	/* m = inv(this')*m */
	void multInvLeftTrans(GeneralMatrix& m) const
		{multInvLeft("T", m);}

	virtual void print() const;
protected:
	void multInvLeft(const char* trans, GeneralMatrix& m) const;
};


class GeneralMatrix {
	friend class ConstGeneralMatrix;
protected:
	Vector data;
	int rows;
	int cols;
	int ld;
public:
	GeneralMatrix(int m, int n)
		: data(m*n), rows(m), cols(n), ld(m) {}
	GeneralMatrix(const double* d, int m, int n)
		: data(d, m*n), rows(m), cols(n), ld(m) {}
	GeneralMatrix(double* d, int m, int n)
		: data(d, m*n), rows(m), cols(n), ld(m) {}
	GeneralMatrix(const GeneralMatrix& m);
	GeneralMatrix(const ConstGeneralMatrix& m);
	GeneralMatrix(const GeneralMatrix&m, const char* dummy); // transpose
	GeneralMatrix(const ConstGeneralMatrix&m, const char* dummy); // transpose
	GeneralMatrix(const GeneralMatrix& m, int i, int j, int nrows, int ncols);
	GeneralMatrix(GeneralMatrix& m, int i, int j, int nrows, int ncols);
	GeneralMatrix(const GeneralMatrix& a, const GeneralMatrix& b);
	virtual ~GeneralMatrix();
	const GeneralMatrix& operator=(const GeneralMatrix& m)
		{data=m.data; rows=m.rows; cols=m.cols; ld=m.ld; return *this;}

	const double& get(int i, int j) const
		{return data[j*ld+i];}
	double& get(int i, int j)
		{return data[j*ld+i];}
	int numRows() const {return rows;}
	int numCols() const {return cols;}
	int getLD() const {return ld;}
	double* base() {return data.base();}
	const double* base() const {return data.base();}
	Vector& getData() {return data;}
	const Vector& getData() const {return data;}

	double getNormInf() const
		{return ConstGeneralMatrix(*this).getNormInf();}
	double getNorm1() const
		{return ConstGeneralMatrix(*this).getNorm1();}

	/* place matrix m to the position (i,j) */
	void place(const ConstGeneralMatrix& m, int i, int j);
	void place(const GeneralMatrix& m, int i, int j)
		{place(ConstGeneralMatrix(m), i, j);}

	/* this = a*b */
	void mult(const ConstGeneralMatrix& a, const ConstGeneralMatrix& b);
	void mult(const GeneralMatrix& a, const GeneralMatrix& b)
		{mult(ConstGeneralMatrix(a), ConstGeneralMatrix(b));}

	/* this = this + a*b */
	void multAndAdd(const ConstGeneralMatrix& a, const ConstGeneralMatrix& b);
	void multAndAdd(const GeneralMatrix& a, const GeneralMatrix& b)
		{multAndAdd(ConstGeneralMatrix(a), ConstGeneralMatrix(b));}

	/* this = this * m */
	void multRight(const ConstGeneralMatrix& m);
	void multRight(const GeneralMatrix& m)
		{multRight(ConstGeneralMatrix(m));}

	/* this = m * this */
	void multLeft(const ConstGeneralMatrix& m);
	void multLeft(const GeneralMatrix& m)
		{multLeft(ConstGeneralMatrix(m));}

	/* this = this * m' */
	void multRightTrans(const ConstGeneralMatrix& m);
	void multRightTrans(const GeneralMatrix& m)
		{multRightTrans(ConstGeneralMatrix(m));}

	/* this = m' * this */
	void multLeftTrans(const ConstGeneralMatrix& m);
	void multLeftTrans(const GeneralMatrix& m)
		{multLeftTrans(ConstGeneralMatrix(m));}

	/* x = x + this*d */
	void multaVec(Vector& x, const ConstVector& d) const
		{ConstGeneralMatrix(*this).multaVec(x, d);}

	/* x = x + this'*d */
	void multaVecTrans(Vector& x, const ConstVector& d) const
		{ConstGeneralMatrix(*this).multaVecTrans(x, d);}

	/* x = x - this*d */
	void multsVec(Vector& x, const ConstVector& d) const
		{ConstGeneralMatrix(*this).multsVec(x, d);}

	/* x = x - this'*d */
	void multsVecTrans(Vector& x, const ConstVector& d) const
		{ConstGeneralMatrix(*this).multsVecTrans(x, d);}

	/* this = zero */
	void zeros();

	/* this = scalar*this */
	void mult(double a);

	/* this = this + scalar*m */
	void add(double a, const ConstGeneralMatrix& m);
	void add(double a, const GeneralMatrix& m)
		{add(a, ConstGeneralMatrix(m));}

	virtual void print() const
		{ConstGeneralMatrix(*this).print();}
private:
	void copy(const ConstGeneralMatrix& m, int ioff = 0, int joff = 0);
	void copy(const GeneralMatrix& m, int ioff = 0, int joff = 0)
		{copy(ConstGeneralMatrix(m), ioff, joff);}

	void gemm(const char* transa, const ConstGeneralMatrix& a,
			  const char* transb, const ConstGeneralMatrix& b,
			  double alpha, double beta);
	void gemm(const char* transa, const GeneralMatrix& a,
			  const char* transb, const GeneralMatrix& b,
			  double alpha, double beta)
		{gemm(transa, ConstGeneralMatrix(a), transb, ConstGeneralMatrix(b),
			  alpha, beta);}

	/* this = this * op(m) (without whole copy of this) */
	void gemm_partial_right(const char* trans, const ConstGeneralMatrix& m,
							double alpha, double beta);
	void gemm_partial_right(const char* trans, const GeneralMatrix& m,
							double alpha, double beta)
		{gemm_partial_right(trans, ConstGeneralMatrix(m), alpha, beta);}

	/* this = op(m) *this (without whole copy of this) */
	void gemm_partial_left(const char* trans, const ConstGeneralMatrix& m,
						   double alpha, double beta);
	void gemm_partial_left(const char* trans, const GeneralMatrix& m,
						   double alpha, double beta)
		{gemm_partial_left(trans, ConstGeneralMatrix(m), alpha, beta);}

	/* number of rows/columns for copy used in gemm_partial_* */
	static int md_length;
};





#endif /* GENERAL_MATRIX_H */


// Local Variables:
// mode:C++
// End:
